/* Copyright (c) 2013 Oberon microsystems, Inc. (Switzerland)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. */

// Originally developed for the book
//   "Getting Started with the Internet of Things", by Cuno Pfister.
//   Copyright 2011 Cuno Pfister, Inc., 978-1-4493-9357-1.
//
// Version 1.1, for the .NET Micro Framework release 4.2.

using System.Diagnostics.Contracts;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace Gsiot.Http.Sockets
{
    // server-side abstraction: listener

    public sealed class SocketStreamListener : IStreamListener
    {
        public bool IsOpen { get; private set; }
        public string LocalHostName { get; private set; }
        public int LocalPort { get; set; }

        // configuration properties specific to sockets (usually Ethernet sockets)
        public int Backlog { get; set; }

        Socket listener;

        static void DebugPrint(string s)
        {
            //Microsoft.SPOT.Debug.Print(s);
        }

        public SocketStreamListener()
        {
            IsOpen = false;
            LocalHostName = null;
            LocalPort = -1;
            Backlog = 0;
        }

        public void Open()
        {
            Contract.Requires(!IsOpen);
            Contract.Requires(LocalPort >= 0);
            Contract.Requires(LocalPort <= 65535);
            if (Backlog <= 0) { Backlog = 4; }

            Contract.Assert(listener == null);
            try
            {
                IPAddress myAdr = IPAddress.GetDefaultLocalAddress();
                while (myAdr == IPAddress.Any)
                {
                    Thread.Sleep(1000);             // wait until local address is set
                    myAdr = IPAddress.GetDefaultLocalAddress();
                }
                LocalHostName = myAdr.ToString();
                DebugPrint("SocketStreamListener: local host is " + LocalHostName);
                listener = new Socket(AddressFamily.InterNetwork,
                                        SocketType.Stream, ProtocolType.Tcp);
                listener.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, true);
                listener.Bind(new IPEndPoint(IPAddress.Any, LocalPort));
                DebugPrint("SocketStreamListener: open listener for incoming stream");
                listener.Listen(Backlog);
                IsOpen = true;
            }
            catch (SocketException e)
            {
                if (listener != null) { Dispose(); }
                listener = null;
                throw new StreamException("socket error " + e.ErrorCode, e);
            }
        }

        public void Dispose()
        {
            if (IsOpen)
            {
                IsOpen = false;
                Contract.Assert(listener != null);
                listener.Close();
                listener = null;
                LocalHostName = null;
            }
        }

        public void Close()
        {
            Dispose();
        }

        public Stream Accept()
        {
            DebugPrint("SocketStreamListener: accepting...");
            try
            {
                Socket s = listener.Accept();
                // Socket.Accept may throw an exception: SocketException,
                // ObjectDisposedException, or InvalidOperationException
                return new NetworkStream(s);
            }
            catch (SocketException e)     // this may happen, treat as expected error
            {
                throw new StreamException("socket error " + e.ErrorCode, e);
            }
        }
    }


    // client-side abstraction: socket

    public sealed class EthernetStreamFactory : IStreamFactory
    {
        public bool IsOpen { get; private set; }
        public string RemoteHostName { get; set; }
        public int RemotePort { get; set; }

        static void DebugPrint(string s)
        {
            //Microsoft.SPOT.Debug.Print(s);
        }

        public EthernetStreamFactory()
        {
            IsOpen = false;
            RemoteHostName = null;
            RemotePort = -1;
            DebugPrint("SocketStreamListener: stream factory created");
        }

        public void Open()
        {
            Contract.Requires(!IsOpen);
            Contract.Requires(RemoteHostName != null);
            Contract.Requires(RemoteHostName.Length > 0);
            Contract.Requires(RemotePort >= 0);
            Contract.Requires(RemotePort <= 65535);

            IsOpen = true;
            DebugPrint("SocketStreamListener: stream factory opened");
        }

        public void Dispose()
        {
            if (IsOpen)
            {
                IsOpen = false;
                RemoteHostName = null;
                DebugPrint("SocketStreamListener: stream factory disposed");
            }
        }

        public void Close()
        {
            Dispose();
        }

        public Stream Connect()
        {
            Contract.Requires(IsOpen);

            // look up relay host's domain name,
            // to find IP address(es)
            IPHostEntry hostEntry = null;
            try
            {
                DebugPrint("SocketStreamListener: resolving " + RemoteHostName + "...");
                hostEntry = Dns.GetHostEntry(RemoteHostName);
                DebugPrint("SocketStreamListener: resolved " + RemoteHostName);
            }
            catch (SocketException)
            {
                DebugPrint("SocketStreamListener: could not resolve DNS name (is DHCP enabled?) " + RemoteHostName);
            }
            // extract a returned address
            try
            {
                IPAddress hostAddress = hostEntry.AddressList[0];
                IPEndPoint remoteEndPoint = new IPEndPoint(hostAddress, RemotePort);

                var socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                socket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, true);
                DebugPrint("SocketStreamListener: connecting to " + RemoteHostName + "...");
                socket.Connect(remoteEndPoint);
                DebugPrint("SocketStreamListener: connected to " + RemoteHostName);
                return new NetworkStream(socket);
            }
            catch (SocketException s)
            {
                Dispose();
                DebugPrint("SocketStreamListener: cannot connect to server, socket exception " + s.ErrorCode);
                throw new StreamException("cannot connect to server", s);
            }
        }
    }
}
